
#ifndef _SDL_test_h
#define _SDL_test_h

#include "SDL.h"
#include "SDL_test_common.h"
#include "SDL_test_font.h"
#include "SDL_test_random.h"
#include "SDL_test_fuzzer.h"
#include "SDL_test_crc32.h"
#include "SDL_test_md5.h"
#include "SDL_test_log.h"
#include "SDL_test_assert.h"
#include "SDL_test_harness.h"
#include "SDL_test_images.h"
#include "SDL_test_compare.h"

#include "begin_code.h"

#ifdef __cplusplus
extern "C" {
#endif

#define SDLTEST_MAX_LOGMESSAGE_LENGTH   3584

#ifdef __cplusplus
}
#endif
#include "close_code.h"

#endif

